"""Work Object."""

from json import loads
from os import environ
from pathlib import Path
from time import time
from typing import Any, Dict, List, Literal, Optional, Union
from warnings import warn

from pydantic import BaseModel, Field, StrictFloat, StrictStr, root_validator
from tenacity import retry
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_random

from chime_frb_api.modules.buckets import Buckets


class Work(BaseModel):
    """The Work Object.

    Example:
        ```python
        from chime_frb_api.workflow.work import Work

        work = Work(
            pipeline="test",
        )
        work.deposit()
        ```

    Args:
        pipeline (str): Name of the pipeline.
        function (str, optional): Name of the function ran as `function(**parameters)`.
        command (List[str], optional): Command to run as `subprocess.run(command)`.
        parameters (Dict[str, Any], optional): Parameters to pass the function.
        results (Dict[str, Any], optional): Results of the work performed, if any.
        plots (List[str], optional): Path to visual data products. These are
            direct paths to the data products or paths relative to work.path.
            If archive is True, these data paths will be automatically archived
            by the Datatrail API.
        products (List[str], optional): Path to the non-human-readable data products
            generated by the pipeline. These are direct paths to the data products
            or paths relative to work.path. If archive is True, these data paths
            will be automatically archived by the Datatrail API.
        path (str, optional): Base path where the pipeline results are stored.
            Defaults to "." or the value of the WORK_PATH environment variable.
            Used by downstream pipelines to locate work products and plots.
        event (List[int], optional): CHIME/FRB Event ID[s].
        tags (List[str], optional): Searchable tags for the work.
            Merged with env WORK_TAGS.
        group (str, optional): CHIME/FRB Working Group. Also sourced from env WORK_GROUP.
        timeout (int, optional): Timeout in seconds for the work to finish.
            Defaults 3600s (1 hr) with range of [1, 86400] (1s-24hrs).
        retries (int, optional): Number of retries before giving up. Defaults to 2.
        priority (int, optional): Priority of the work. Defaults to 3. Range [1, 5].
        site (str, optional): Site where the work was performed. Defaults to "local".
        user (str, optional): User who performed the work. Defaults to None.
        archive (bool, optional): Whether to archive the work products and plots.

    Raises:
        pydantic.ValidationError: If any of the arguments are of wrong type or value.

    Returns:
        Work: Work Object.
    """

    class Config:
        """Pydantic Config."""

        validate_all = True
        validate_assignment = True

    ###########################################################################
    # Required Attributes. Set by user.
    ###########################################################################
    pipeline: StrictStr = Field(
        ...,
        min_length=1,
        description="Name of the pipeline.",
        example="example-pipeline",
    )
    ###########################################################################
    # Optional attributes, might be provided by the user.
    ###########################################################################
    function: str = Field(
        default=None,
        description="""
        Name of the function to run as `function(**parameters)`.
        Only either `function` or `command` can be provided.
        """,
        example="requests.get",
    )
    parameters: Optional[Dict[str, Any]] = Field(
        default=None,
        description="""
        Parameters to pass the pipeline function.
        """,
        example={"event_number": 9385707},
    )
    command: List[str] = Field(
        default=None,
        description="""
        Command to run as `subprocess.run(command)`.
        Only either `function` or `command` can be provided.
        """,
        example=["python", "example.py", "--example", "example"],
    )
    results: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Results of the work performed, if any.",
        example={"dm": 100.0, "snr": 10.0},
    )
    plots: Optional[List[StrictStr]] = Field(
        default=None,
        description="""
        Name of visual data products generated by the pipeline.
        These are direct paths to the data products or paths relative to work.path.
        If archive is True, these data paths will be automatically archived by the
        Datatrail API.
        """,
        example=["waterfall.png", "/arc/projects/chimefrb/9385707/9385707.png"],
    )
    products: Optional[List[StrictStr]] = Field(
        default=None,
        description="""
        Name of the non-human-readable data products generated by the pipeline.
        These are direct paths to the data products or paths relative to work.path.
        If archive is True, these data paths will be automatically archived by the
        Datatrail API.
        """,
        example=["spectra.h5", "dm_vs_time.png"],
    )
    path: Optional[str] = Field(
        default=environ.get("WORK_PATH", "."),
        description="""
        Base path where the pipeline results are stored.
        Defaults to "." or the value of the WORK_PATH environment variable.
        Used by downstream pipelines to locate work products and plots.
        """,
        example="/arc/projects/chimefrb/",
    )
    event: Optional[List[int]] = Field(
        default=None,
        description="CHIME/FRB Event ID[s] the work was performed against.",
        example=[9385707, 9385708],
    )
    tags: Optional[List[str]] = Field(
        default=None,
        description="""
        Searchable tags for the work. Merged with values from env WORK_TAGS.
        """,
        example=["dm-analysis"],
    )
    group: Optional[StrictStr] = Field(
        default=environ.get("WORK_GROUP", None),
        description="""
        CHIME/FRB Working Group. Can be sourced from env WORK_GROUP.
        """,
        example="properties-wg",
    )
    timeout: int = Field(
        default=3600,
        ge=1,
        le=86400,
        description="""
        Timeout in seconds for the work to finish.
        Defaults 3600s (1 hr) with range of [1, 86400] (1s-24hrs).
        """,
        example=7200,
    )
    retries: int = Field(
        default=2,
        lt=6,
        description="Number of retries before giving up. Defaults to 2.",
        example=4,
    )
    priority: int = Field(
        default=3,
        ge=1,
        le=5,
        description="Priority of the work. Defaults to 3.",
        example=1,
    )
    site: Literal[
        "chime", "allenby", "kko", "gbo", "hco", "canfar", "cedar", "local", "aro"
    ] = Field(
        default=environ.get("WORK_SITE", "local"),
        description="Site where the work will be performed.",
        example="chime",
    )
    user: Optional[StrictStr] = Field(
        default=None, description="User ID who created the work.", example="shiny"
    )
    archive: bool = Field(
        default=True,
        description="""
        Whether or not to archive the work object, results, plots & products.
        """,
        example=True,
    )
    config: Optional[str] = Field(
        default=None,
        description="UUID of parent pipeline configuration.",
        example="e6b7c8d9-e0a1-2b3c-4d5e-6f7a8b9c0d1e",
    )

    # Deprecated Attributes to be removed in future versions.
    precursors: Optional[List[Dict[StrictStr, StrictStr]]] = Field(
        default=None,
        deprecated=True,
        description="This field has been deprecated.",
    )

    ###########################################################################
    # Automaticaly set attributes
    ###########################################################################
    id: Optional[StrictStr] = Field(
        default=None, description="Work ID created by the database."
    )
    creation: Optional[StrictFloat] = Field(
        default=None, description="Unix timestamp of when the work was created."
    )
    start: Optional[StrictFloat] = Field(
        default=None,
        description="Unix timestamp when the work was started, reset at each attempt.",
    )
    stop: Optional[StrictFloat] = Field(
        default=None,
        description="Unix timestamp when the work was stopped, reset at each attempt.",
    )
    attempt: int = Field(
        default=0, ge=0, description="Attempt number at performing the work."
    )
    status: Literal["created", "queued", "running", "success", "failure"] = Field(
        default="created", description="Status of the work."
    )
    ###########################################################################
    # Attribute setters for the work attributes
    ###########################################################################

    @root_validator
    def post_init(cls, values: Dict[str, Any]):
        """Initialize work attributes after validation."""
        # Set creation time if not already set
        if values.get("creation") is None:
            values["creation"] = time()
        # Update tags from environment variable WORK_TAGS
        if environ.get("WORK_TAGS"):
            env_tags: List[str] = str(environ.get("WORK_TAGS")).split(",")
            # If tags are already set, append the new ones
            if values.get("tags"):
                values["tags"] = values["tags"] + env_tags
            else:
                values["tags"] = env_tags
            # Remove duplicates
            values["tags"] = list(set(values["tags"]))

        # Check if both command and function are set
        if values.get("command") and values.get("function"):
            raise ValueError("command and function cannot be set together.")

        # Check path exists and is a directory
        if values.get("path"):
            path = Path(values.get("path"))  # type: ignore
            assert path.exists(), f"{values.get('path')} does not exist."
            assert path.is_dir(), f"{values.get('path')} is not a directory."

        # Display deprecation warning for precursors & config
        if values.get("precursors"):
            warn(
                """\n
                The `precursors` attribute has been deprecated.
                It will be removed in chime-frb-api v3.0.0.
                Please remove it from your code.\n""",
                DeprecationWarning,
                stacklevel=2,
            )
            values["precursors"] = None
            values["config"] = None

        if values.get("site") == "allenby":
            warn(
                """\n
                The `allenby` site name has been deprecated and will be
                removed from chime-frb-api by summer 2023 release.
                Please use site name `kko` instead.\n
                """,
            )
        return values

    ###########################################################################
    # Work methods
    ###########################################################################

    @property
    def payload(self) -> Dict[str, Any]:
        """Return the dictioanary representation of the work.

        Returns:
            Dict[str, Any]: The payload of the work.
            Non-instanced attributes are excluded from the payload.
        """
        payload: Dict[str, Any] = self.dict()
        return payload

    @classmethod
    def from_json(cls, json_str: str) -> "Work":
        """Create a work from a json string.

        Args:
            json_str (str): The json string.

        Returns:
            Work: The work.
        """
        return cls(**loads(json_str))

    @classmethod
    def from_dict(cls, payload: Dict[str, Any]) -> "Work":
        """Create a work from a dictionary.

        Args:
            payload (Dict[str, Any]): The dictionary.

        Returns:
            Work: The work.
        """
        return cls(**payload)

    ###########################################################################
    # HTTP Methods
    ###########################################################################

    @classmethod
    def withdraw(
        cls,
        pipeline: str,
        event: Optional[List[int]] = None,
        site: Optional[str] = None,
        priority: Optional[int] = None,
        user: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Optional["Work"]:
        """Withdraw work from the buckets backend.

        Args:
            pipeline (str): Name of the pipeline.
            **kwargs (Dict[str, Any]): Keyword arguments for the Buckets API.

        Returns:
            Work: Work object.
        """
        buckets = Buckets(**kwargs)  # type: ignore
        payload = buckets.withdraw(
            pipeline=pipeline, event=event, site=site, priority=priority, user=user
        )
        if payload:
            return cls.from_dict(payload)
        return None

    @retry(wait=wait_random(min=0.5, max=1.5), stop=(stop_after_delay(30)))
    def deposit(
        self, return_ids: bool = False, **kwargs: Dict[str, Any]
    ) -> Union[bool, List[str]]:
        """Deposit work to the buckets backend.

        Args:
            **kwargs (Dict[str, Any]): Keyword arguments for the Buckets API.

        Returns:
            bool: True if successful, False otherwise.
        """
        buckets = Buckets(**kwargs)  # type: ignore
        return buckets.deposit(works=[self.payload], return_ids=return_ids)

    @retry(wait=wait_random(min=0.5, max=1.5), stop=(stop_after_delay(30)))
    def update(self, **kwargs: Dict[str, Any]) -> bool:
        """Update work in the buckets backend.

        Args:
            **kwargs (Dict[str, Any]): Keyword arguments for the Buckets API.

        Returns:
            bool: True if successful, False otherwise.
        """
        buckets = Buckets(**kwargs)  # type: ignore
        return buckets.update([self.payload])

    @retry(wait=wait_random(min=0.5, max=1.5), stop=(stop_after_delay(30)))
    def delete(self, **kwargs: Dict[str, Any]) -> bool:
        """Delete work from the buckets backend.

        Args:
            ids (List[str]): List of ids to delete.

        Returns:
            bool: True if successful, False otherwise.
        """
        buckets = Buckets(**kwargs)  # type: ignore
        return buckets.delete_ids([str(self.id)])
